import sys, subprocess, Image, time, os

class SplitterBase():
    """This class will be the parent class that all splitters inherit from"""

    distance = 100 #How far away a color has to be from another color in order for it to not be considered the same color
    number_of_words = 10
    sample_rate = 11025.0
    tempdir = "temp/"
    def __init__(self):
        return

    def split_captcha(self, filename):
        """Given the mp3 of an audio captcha, this class splits the file up into the individual words"""
        print "This method has not been implemented yet."
        sys.exit()

    def get_spectrogram(self, filename, highpass_filter = False, noisegain = False, monocolor = False, width = -1):
        """Get a spectrogram using sox"""
        options = ["sox", self.filename_prefix + filename + self.filename_suffix, "-c", "1", "-n" ]
        if highpass_filter:
            options.extend(["highpass", "1000"])
        if noisegain:
            options.extend(["compand", ".25,.25", "-20.7,-inf", "0", "0", "0"])
        options.extend(["spectrogram", "-o", self.filename_prefix + filename + ".png"])
        if monocolor:
            options.append("-m")
        if width != -1:
            options.extend(["-x", str(width)])
        p = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate()

    def get_duration(self, audiofile, in_seconds = False):
        """Gets the number of frames for the specified audio file"""
        p = subprocess.Popen(["sox", audiofile, "-c", "1", "-n", "stat"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output = p.communicate()[1]
        if in_seconds:
            sampleLine = output.split("\n")[1]
        else:
            sampleLine = output.split("\n")[0]
        return float(sampleLine.split(":")[1].replace(" ", ""))

    def within_distance(self, color1, color2):
        """Determines whether or not a color is close enough to another color to be considered the same color"""
        if type(color1) is tuple and type(color2) is tuple:
            for i in range(0,3):
                if abs(color1[i] - color2[i]) > self.distance:
                    return False
            return True

    def remove_white_border(self, image):
        """Removes the white border from sox's spectrogram (there really should be a sox option for this)"""
        pixels = image.load()
        sizex,sizey = image.size
        centerx = int(sizex / 2)
        centery = int(sizey / 2)
        border_color = (127,127,127)
        top = left = right = bottom = None

        for x in range(centerx, 0, -1): #Start at the center and move in each direction to look for the border color
            if self.within_distance(pixels[x,centery],border_color):
                left = x
                break
        for x in range(centerx, sizex):
            if self.within_distance(pixels[x,centery],border_color):
                right = x
                break
        for y in range(centery, 0, -1):
            if self.within_distance(pixels[centerx,y],border_color):
                top = y
                break
        for y in range(centery, sizey):
            if self.within_distance(pixels[centerx,y],border_color):
                bottom = y
                break

        if left == None or right == None or top == None or bottom == None:
            self.logger.warning("Could not remove the spectrogram border, assuming the image has already been cropped")
            return image

        box = (left + 1, top + 1, right - 1, bottom - 1)
        self.logger.debug("Removing spectrogram border by cropping image to " + str(box))
        return image.crop(box), box

    def highpass(self, filename, frequency = 1000):
        tempfile = self.tempdir + "captcha_highpass_" + str(os.getpid()) + filename[-4:]
        options = ["sox", filename, "-c", "1", tempfile, "highpass", str(frequency)]
        p = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate()
        os.rename(tempfile, filename)

    def convert_file(self, filename, old, new, new_sample_rate = -1, new_filename = ""):
        if new_filename == "":
            new_filename = filename[:-4] + new
        self.logger.debug("Converting %s to %s", filename[:-4] + old, new_filename)
        options = ["sox", filename[:-4] + old, "-c", "1", new_filename]
        if new_sample_rate != -1:
            options.extend(["rate", str(new_sample_rate)])
        p = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate() 
    
    def get_noise_profile(self, filename, profile_name, noise_start = -1, noise_end = -1): 
        noise_profile_start = time.time()
        self.logger.debug("Getting noise profile of %s", filename)
        options = [ "sox", filename, "-n"]
        if noise_start != -1:
            options.append("trim")
            options.append(str(noise_start))
            if noise_end != -1:
                options.append(str(noise_end-noise_start))
            else:
                options.append("99999") #Just give it the rest of the file if no end was specified
        options.append("noiseprof")
        options.append(profile_name)
        profile_process = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) #TODO change this to use a static noise profile
        profile_process.communicate()
        self.logger.debug("Noise profiling took %f seconds",time.time() - noise_profile_start)
 
    def remove_noise(self, filename, profile_name, amount = 0.5, extension = ".mp3"):
        noise_reduction_start = time.time()
    
        self.logger.debug("Reducing noise profile with the threshold of %f", amount)
        options = [ "sox", filename, filename + "_clean" + extension, "noisered", profile_name, str(amount) ]
        clean_process = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        clean_process.communicate()
        os.rename(filename + "_clean" + extension, filename[:-4] + extension)
        self.logger.debug("Noise reduction took %f seconds",time.time() - noise_reduction_start)

    def split_wav_file(self, column_xs, sizex, filename, suffix = "", columnx_unit = "pixel", output_name = [], soundsizex = -1):
        """Splits the corresponding wav file based on the columns found in the spectrogram.  The column x start and x end should be
             passed in a list of tuples in column_xs and they should not overlap.  The width of the image should be passed in sizex, and
             optionally a naming suffix for the split up wav files can be passed in suffix."""
        fullfile = self.filename_prefix + filename + self.filename_suffix
        if soundsizex == -1:
            soundsizex = self.get_duration(fullfile)
        sizex = float(sizex)
        pos = 0.0
        sound_columns = []
        processes = []
        for i in range(0,len(column_xs)):
            #Sox has the ability to split multiple parts at once, but it doesn't work (but I couldn't get it to work)
            options = ["sox", fullfile]
        
            if len(output_name) != 0:
                options.append(output_name[i] + ".wav") #Allow a custom name if there is only one piece
            else:
                options.append(self.filename_prefix + filename + "_" + str(i) + suffix + ".wav")
            options.append("trim")

            if columnx_unit == "pixel":
                skip = float(column_xs[i][0]) * (soundsizex/sizex) * (1.0 / self.sample_rate) #The number of seconds to skip
                length = float(column_xs[i][1] - column_xs[i][0] + 1.0) * (soundsizex / sizex) * (1.0 / self.sample_rate) #The number of seconds to read
            elif columnx_unit == "sample":
                skip = column_xs[i][0] / self.sample_rate
                length = (column_xs[i][1] - column_xs[i][0]) / self.sample_rate
            else: #seconds
                skip = column_xs[i][0]
                if skip < 0:
                    skip = 0
                length = column_xs[i][1] - column_xs[i][0]

            options.append(str(skip))
            options.append(str(length))
            processes.append(subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE))
            sound_columns.append([skip, skip + length])

        for i in range(0, len(processes)):
            processes[i].communicate() #Wait for the last sox process to end
            if len(output_name) != 0:
                os.rename(output_name[i] + ".wav", output_name[i]) #Sox can't overwrite the file it's working on, 

        self.logger.debug("Sound Column x-coordinates are: " + str(sound_columns))

    def trim_image(self, filename = "", threshold = 42000):
        """Takes in a monocolor spectrogram with no border and trims it"""
        old = self.filename_suffix
        self.filename_suffix = '.wav'
        self.get_spectrogram(filename)
        spectrogram = self.filename_prefix + filename + ".png"
        color_image = Image.open(spectrogram).convert("RGB")
        color_image,box = self.remove_white_border(color_image)
        self.get_spectrogram(filename, monocolor = True)
        image = Image.open(spectrogram).convert("RGB")
        image = image.crop(box)
        sizex, sizey = image.size
        pixels = image.load()
        self.filename_suffix = old

        row_counts = {}
        std_dev = {}
        for x in range(0, sizex):
            row_counts[x] = 0 
            values = []
            for y in range(0, sizey):
                row_counts[x] += pixels[x,y][0]
                values.append(pixels[x,y][0])
            std_dev[x] = numpy.std(values)
        left = 0
        while left < sizex and row_counts[left] < threshold:
            left += 1
        right = sizex-1
        while right > 0 and row_counts[right] < threshold:
            right -= 1

        if left <= right: #If we've traversed the entire image, we've obviously failed.  Just don't crop
            self.logger.debug("Cropping image from (0,%d) to (%d,%d)",sizex, left, right)
            color_image = color_image.crop((left, 0, right, sizey))
        else:
            self.logger.debug("Could not trim image correctly, got borders as (%d,%d)", left, right)
            left = 0
            right = sizex
    
        return color_image, sizex, (left, right)

