import logging, argparse, random, os, sys, operator, shutil, subprocess, hashlib, threading
import nnsolver, splot
import lowspectro_splitter, fuckit_splitter
import demo_downloader, paypal_downloader

#########Defaults############
default_bitmask = 24
default_hidden = 1536
default_iterations = 100
#############################


class Trainer():
    """This class will use genetic algorithm to train new neural networks"""

    classname = 'Trainer'
    tempdir = 'temp/trainer/'
    processes_at_once = 50
    captchas_per_round = 100 #250
    stagnant_rounds = 5 #Number of rounds in a row with no improvment before we get more captchas for this round
    input_sizes = { 8 : "2048", 24 : "2049", 32 : "6144" } #Add more if you run it with others

    def __init__(self, debug, downloaders, splitters, test_dir, download_dir, theta_files, ymaps, sample_dirs,
                         bitmasks, hiddens, iterations, filename_suffix, funr = ""):
        self.logger = logging.getLogger(self.classname)
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)

        self.round_num = [0] * len(theta_files)
        self.debug = debug
        self.downloaders = downloaders
        self.test_dir = test_dir
        self.download_dir = download_dir

        self.theta_files = theta_files
        self.ymaps = ymaps
        self.bitmasks = bitmasks
        self.hiddens = hiddens
        self.iterations = iterations
        self.filename_suffix = filename_suffix
        self.funr = funr
        
        if not os.path.exists(self.tempdir):
            os.mkdir(self.tempdir)
        
        for i in range(0,len(sample_dirs)):
            self.cleanup(i)
            inputs_dir = self.tempdir + "inputs_set_" + str(i) + "_" + str(self.round_num[i]) + "/"
            if not os.path.exists(inputs_dir):
                os.mkdir(inputs_dir)
            self.generate_inputs(sample_dirs[i], inputs_dir, self.bitmasks[i]) 
    
        self.theta_files = {}
        self.solvers = {}
        for i in range(0,len(theta_files)):
            self.new_round(i, theta_files[i])

        self.splitters = splitters

    def run(self):
        current_accuracies = self.test(self.theta_files, self.ymaps, self.bitmasks)
        self.download_captchas(self.theta_files.keys(), self.captchas_per_round)
        self.split_captchas(self.theta_files.keys())

        bad_rounds = 0
        while True:
            self.solve_and_sort_captchas()
            new_thetas = self.train(percent = random.random() * .6 + .8)
            new_accuracies = self.test(new_thetas, self.ymaps, self.bitmasks)
            better_thetas = []
            for i in range(0, len(new_accuracies)):
                if new_accuracies[i] > current_accuracies[i]:
                    self.logger.info("Found theta files with better accuracy for sibling %d: Previous accuracy = %f, Current accuracy = %f", i, current_accuracies[i], new_accuracies[i])
                    current_accuracies[i] = new_accuracies[i]
                    trainer.new_round(i, new_thetas[i])
                    better_thetas.append(i)
                else:
                    self.logger.info("New theta had worse accuracy for sibling %d: Previous accuracy = %f, Current accuracy = %f", i, current_accuracies[i], new_accuracies[i])
                    self.cleanup(i)

            if len(better_thetas) == 0 and bad_rounds >= self.stagnant_rounds:
                self.logger.info("We've gone %d rounds without a better theta, getting new captchas to prevent stagnating", bad_rounds)
                better_thetas = self.theta_files.keys()

            if len(better_thetas) != 0:
                bad_rounds = 0
                self.download_captchas(better_thetas, self.captchas_per_round)
                self.split_captchas(better_thetas)
            else:
                bad_rounds += 1

    def cleanup(self, index):
        delete_files = []
        delete_files.append(self.tempdir + "inputs_" + str(index) + "_" + str(self.round_num[index]) + "_x.csv")
        delete_files.append(self.tempdir + "inputs_" + str(index) + "_" + str(self.round_num[index]) + "_y.csv")
        delete_files.append(self.tempdir + "inputs_" + str(index) + "_" + str(self.round_num[index]) + "_ymap.txt")
        delete_files.append(self.tempdir + "theta_values_auto_" + str(self.round_num[index]) + "_" + str(self.bitmasks[index]) + "_" + 
                                str(self.hiddens[index]) + "_" + str(self.iterations[index]) + ".mat")
        delete_files.append(self.tempdir + "solved_set_" + str(index) + "_" + str(self.round_num[index]) + "/")
        delete_files.append(self.tempdir + "inputs_set_" + str(index) + "_" + str(self.round_num[index]) + "/")
       
        for delete_file in delete_files:
            if os.path.exists(delete_file):
                if os.path.isdir(delete_file):
                    shutil.rmtree(delete_file)
                else:
                    os.remove(delete_file)

    def new_round(self, index, theta_file):
        self.theta_files[index] = theta_file
        self.solvers[index] = nnsolver.NnSolver(self.debug, self.bitmasks[index], theta_matrix_files = [ theta_file ], ymapping_files = [ self.ymaps[index] ]) 
        self.round_num[index] += 1

    def download_captchas(self, theta_indices, howmany):
        for downloader in self.downloaders:
            downloader.tempdir = self.download_dir

        total = howmany * len(theta_indices)
        total -= len(os.listdir(self.download_dir))
        if total < 0:
            total = 0

        round_str = ""
        for round_num in self.round_num:
            round_str += str(round_num)

        i = 0
        self.logger.info("Downloading %d captchas to %s", total, self.download_dir)
        while i < total:
            filename = "captcha_round_" + round_str + "_" + str(i)
            current_downloader = self.downloaders[random.randint(0,len(self.downloaders)-1)]
            challenge, url = current_downloader.get_audio_url_and_challenge()
            if current_downloader.get_audio(url, filename):
                i += 1
       
        captchafiles = [] 
        allfiles = os.listdir(self.download_dir)
        for filename in allfiles:
            if filename.endswith(self.filename_suffix):
                captchafiles.append(filename)
            else:
                os.remove(self.download_dir + "/" + filename)

        for i in range(0, len(theta_indices)):
            index = theta_indices[i]
            captcha_dir = self.tempdir + "captcha_set_" + str(index) + "_" + str(self.round_num[index]) + "/"
            if not os.path.exists(captcha_dir):
                os.mkdir(captcha_dir)

            for j in range(0, howmany):
                filename = captchafiles.pop(random.randint(0, len(captchafiles) - 1))
                print "Moving",self.download_dir + "/" + filename,"to",captcha_dir + filename
                os.rename(self.download_dir + "/" + filename, captcha_dir + filename) 

    def split_captchas(self, theta_indices):
        for theta_index in theta_indices:
            captcha_dir = self.tempdir + "captcha_set_" + str(theta_index) + "_" + str(self.round_num[theta_index]) + "/"
            split_dir = self.tempdir + "split_set_" + str(theta_index) + "_" + str(self.round_num[theta_index]) + "/"
            if not os.path.exists(split_dir):
                os.mkdir(split_dir)
        
            captchas = os.listdir(captcha_dir)
            filelist = []
            for captcha in captchas:
                if captcha.endswith(self.filename_suffix):    
                    captcha = captcha.replace(self.filename_suffix,"")
                    filelist.append(captcha)
            self.logger.info("Splitting %d captchas and saving to %s", len(filelist), split_dir)

            count = 0
            while count < len(filelist):
                threads = []
                for i in range(count, count + self.processes_at_once):
                    if i < len(filelist):
                        splitter = self.splitters[i - count]
                        splitter.filename_prefix = captcha_dir
                        thread = threading.Thread(target = self.split_single_captcha, args = (splitter, filelist[i], captcha_dir, split_dir))
                        thread.start()
                        threads.append(thread)

                for i in range(0, len(threads)):
                    thread.join()
                count += self.processes_at_once

    def split_single_captcha(self, splitter, captcha, captcha_dir, split_dir):
        num_words = splitter.split_captcha(captcha, [ "audio" ])
        for i in range(0, num_words):
            os.rename(captcha_dir + captcha + "_" + str(i) + ".wav", split_dir + captcha + "_" + str(i) + ".wav")
        
    def solve_and_sort_captchas(self):
        for solver_index in range(0, len(self.solvers)):
            threshold = (random.random() * 10 + 60)/100 #Randomly between 10 and 70 
            split_dir = self.tempdir + "split_set_" + str(solver_index) + "_" + str(self.round_num[solver_index]) + "/"
            solved_dir = self.tempdir + "solved_set_" + str(solver_index) + "_" + str(self.round_num[solver_index]) + "/"
            if not os.path.exists(solved_dir):
                os.mkdir(solved_dir)
            
            wordfiles = os.listdir(split_dir)
            filelist = []
            self.logger.info("Solving %d words and saving words with certainity better than %f to %s", len(wordfiles), threshold, solved_dir)
            for wordfile in wordfiles: #It'll be faster if we start solving all the words, so they're more parallaled
                if wordfile.endswith(".wav"):
                    filelist.append(split_dir + wordfile)
            
            count = 0
            while count < len(filelist):
                files = {}
                for i in range(count, count + self.processes_at_once):
                    if i < len(filelist):
                        index = solver_index
                        while index == solver_index:
                            index = random.randint(0, len(self.solvers)-1)

                        if not index in files:
                            files[index] = []
                        files[index].append(filelist[i])

                        solver = self.solvers[index]
                        solver.filename_prefix = split_dir
                        solver.start_solving_word(filelist[i])

                for index, solved_files in files.items():
                    solver = self.solvers[index]
                    for i in range(0, len(solved_files)):
                        answer_matrix = solver.finish_solving_word(i)
                        word,certainity = max(answer_matrix.iteritems(), key=operator.itemgetter(1))
                        self.logger.debug("Sibling %d said %s is %s with %f certainity", index, solved_files[i], word, certainity)
                        if certainity > threshold:
                            dest_dir = solved_dir + "/" + word + "/"
                            self.logger.debug("Copying %s to %s", solved_files[i], dest_dir)
                            if not os.path.exists(dest_dir):
                                os.mkdir(dest_dir)
                            shutil.copy(solved_files[i], dest_dir)
                    solver.reset_solver()
                count += self.processes_at_once

    def generate_inputs(self, input_dir, output_dir, bitmask):
        self.logger.info("Generating inputs for %s and saving in %s", input_dir, output_dir)
        word_dirs = os.listdir(input_dir)
        samples_are_csvs = False
        filelist = []
        for word_dir in word_dirs: #Again, it's faster to start them all at once so they parallel
            if word_dir.endswith("csv"): 
                samples_are_csvs = True
            else:
                word_files = os.listdir(input_dir + word_dir)
                for word_file in word_files:
                    filelist.append(word_dir + "/" + word_file)

        if samples_are_csvs: #If the user gave us a directory with csv's in it then we don't have to regenerate the words
            for word_csv in word_dirs:
                shutil.copy(input_dir + "/" + word_csv, output_dir + "/" + word_csv)
        else:
            wordfiles = {}
            count = 0
            self.logger.info("Starting sox processes %d at a time", self.processes_at_once)
            while count < len(filelist):
                processes = {}
                for i in range(count, count + self.processes_at_once):
                    if i < len(filelist):
                        options = [ "sox", input_dir + filelist[i], "-n", "stat", "-s", "1", "-freq" ]
                        processes[filelist[i]] = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                for filepath, process in processes.items():
                    word = filepath.split("/")[0]
                    if not word in wordfiles:
                        wordfiles[word] = open(output_dir + "/" + word + ".csv", "w")
                    dat = process.communicate()[1].split("\n")
                    splotter = splot.Splot(method = bitmask, lines = dat)
                    output = splotter.extract_data()
                    wordfiles[word].write(output + "\n")
                count += self.processes_at_once

            for wordfile in wordfiles.values():
                wordfile.close()

    def combine_sets(self, old_set, new_set, percent, threshold, solver_index, just_replace = False, best_of_both = True):
        """Combines this round's new words with the last rounds
            old_set = the directory with the previous rounds csv files in it
            new_set = the directory with this rounds csv files in it
            percent = If specified, this will save the given percent of the best words from the previous round, otherwise specify None
            threshold = If specified, this will save the words from the previous round with certainity above the specified threshold, otherwise specify None
            just_replace = Whether we should keep enough old words so that we don't go down in number of words"""
        self.logger.info("Combining inputs for solver %d in %s into %s with percent %f", solver_index, old_set, new_set, percent)
        wordfiles = os.listdir(old_set)
        for wordfile in wordfiles:
            oldfile = open(old_set + wordfile, "r")
            oldlines = oldfile.readlines()
            oldfile.close()

            if os.path.exists(new_set + wordfile):
                newfile = open(new_set + wordfile, "r")
                newlines = newfile.readlines()
                newfile.close()

                word = wordfile.replace(".csv","")
                certainities = {}
                if not best_of_both:
                    newfile = open(new_set + wordfile, "a")
                    for i in range(0, len(oldlines)):
                        index = solver_index
                        while index == solver_index:
                            index = random.randint(0, len(self.solvers)-1)

                        solver = self.solvers[index]
                        nninputs = oldlines[i].replace("\n", "")
                        answer_matrix = solver.nn_predict_word(nninputs)
                        certainities[i] = answer_matrix[word]
                else:
                    newfile = open(new_set + wordfile, "w")
                    oldlines.extend(newlines)
                    for i in range(0, len(oldlines)):
                        index = solver_index
                        while index == solver_index:
                            index = random.randint(0, len(self.solvers)-1)

                        solver = self.solvers[index]
                        nninputs = oldlines[i].replace("\n", "")
                        answer_matrix = solver.nn_predict_word(nninputs)
                        certainities[i] = answer_matrix[word]
                    
                certainities = sorted(certainities.iteritems(), key=operator.itemgetter(1)) 
                certainities.reverse() #Now the captcha pieces are sorted on certainty (highest first)

                line_count = 0
                total_words = len(certainities)
                if just_replace:
                    total_words -= len(newlines)
                elif percent != None:
                    total_words = len(oldlines) * percent 

                if threshold != None:
                    self.logger.debug("Solver %d getting %s's with certainity above %f", solver_index, word, threshold)
                else:
                    self.logger.debug("Solver %d set needs %d old %s's", solver_index, total_words, word)

                for line_index, certainity in certainities:
                    if (threshold != None and certainity > threshold) or ((just_replace or percent != None) and line_count < total_words):
                        self.logger.debug("Solver %d set: Adding %s line with certainity %f", solver_index, word, certainity)
                        newfile.write(oldlines[line_index])
                        line_count += 1

                newfile.close()
            else: #If we have no new ones for this word, then just copy over the old ones
                shutil.copy(old_set + wordfile, new_set + wordfile)
       
    def combine_csvs(self, inputs_dir, outputx, outputy, outputymap):
        """Takes a directory full of csv files and combines them into a single x training file, y training file, and ymap file"""
        self.logger.info("Combining inputs in %s into %s", inputs_dir, outputx)
        input_files = os.listdir(inputs_dir)
        input_files.sort()
        xfile = open(outputx, "w")
        yfile = open(outputy, "w")
        ymapfile = open(outputymap, "w")
        first = True
        wordcount = 1
        for input_file in input_files:
            if input_file.endswith(".csv"):
                wordfile = open(inputs_dir + input_file, "r")
                lines = wordfile.readlines()
                wordfile.close()
                for line in lines:
                    xfile.write(line)
                    if not first:
                        yfile.write("\n")
                    else:
                        first = False
                    yfile.write(str(wordcount))
                if wordcount != 1:
                    ymapfile.write("\n")
                ymapfile.write(str(wordcount) + ", " + input_file.replace(".csv","")) 
                wordcount += 1
        xfile.close()
        yfile.close()
        ymapfile.close()

    def train(self, percent = None, threshold = None):
        processes = []
        output_thetas = []
        for i in range(0, len(self.theta_files)):
            inputsx = self.tempdir + "inputs_" + str(i) + "_" + str(self.round_num[i]) + "_x.csv"
            inputsy = self.tempdir + "inputs_" + str(i) + "_" + str(self.round_num[i]) + "_y.csv"
            inputsymap = self.tempdir + "inputs_" + str(i) + "_" + str(self.round_num[i]) + "_ymap.txt"
            output_theta = (self.tempdir + "theta_values_auto_" + str(i) + "_" + str(self.round_num[i]) + "_" + 
                             str(self.bitmasks[i]) + "_" + str(self.hiddens[i]) + "_" + str(self.iterations[i]) + ".mat")
            output_thetas.append(output_theta)
            solved_dir = self.tempdir + "solved_set_" + str(i) + "_" + str(self.round_num[i]) + "/"
            inputs_dir = self.tempdir + "inputs_set_" + str(i) + "_" + str(self.round_num[i]) + "/"
            prev_inputs_dir = self.tempdir + "inputs_set_" + str(i) + "_" + str(self.round_num[i] - 1) + "/"

            if not os.path.exists(inputs_dir):
                os.mkdir(inputs_dir)

            self.generate_inputs(solved_dir, inputs_dir, self.bitmasks[i]) #Generate the inputs for this round 
            self.combine_sets(prev_inputs_dir, inputs_dir, percent, threshold, i, False) #Combine the last set of inputs and this set
            self.combine_csvs(inputs_dir, inputsx, inputsy, inputsymap) 
             
            self.logger.info("Training %s with %s as inputs", output_theta, inputsx)
            options = [ "octave", "-q", "-H", "-p", "image-neural-network/", "image-neural-network/trainNN3.m", str(self.input_sizes[self.bitmasks[i]]),
                        str(self.hiddens[i]), str(self.iterations[i]), "theta_values_most", "99999999", inputsx, inputsy, output_theta ]
            processes.append(subprocess.Popen(options, stdin = subprocess.PIPE, stdout = subprocess.PIPE))

        for i in range(0, len(processes)):
            processes[i].communicate() #Wait until they're finished 
        
        return output_thetas
    
    def test(self, theta_files, ymaps, bitmasks):
        """Tests the theta files using the offline.py script.  At some point I should just call the relative methods instead of subprocessing to offline.py
            but for now this is good enough"""
        processes = []
        for i in range(0, len(theta_files)):
            self.logger.info("Testing %s with captchas in %s", theta_files[i], self.test_dir)
            options = [ "python", "offline.py", self.test_dir, "-fu" ]
            if self.funr != "":
                options.extend(["-funr", self.funr])
            if self.filename_suffix != "":
                options.extend(["-ft", self.filename_suffix.replace(".", "")])
            options.extend(["-s", "1", "neural", str(bitmasks[i]), theta_files[i], self.input_sizes[bitmasks[i]], ymaps[i]]) 
            self.logger.debug("Running %s", " ".join(options))
            processes.append(subprocess.Popen(options, stdin = subprocess.PIPE, stdout = subprocess.PIPE, stderr = subprocess.PIPE))

        accuracies = []
        for i in range(0, len(processes)):
            output, error = processes[i].communicate()
            accuracy = float(output.split("\n")[-2])
            self.logger.info("%s has accuracy %f", theta_files[i], accuracy)
            accuracies.append(accuracy)
        return accuracies
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="This script will train new networks using genetic algorithms")
    parser.add_argument('test_dir', type=str, help='The directory with the testing set')
    parser.add_argument('download_dir', type=str, help='The directory to download captchas to.  If there are captchas there at start, ' + 
                            'they will be used before downloading.  THESE FILES WILL BE MOVED, ENSURE YOU HAVE A SECOND COPY OF THESE.')

    parser.add_argument('-sample_dir',  type=str, nargs='*', 
                        help='The directory with the initial sample sets (the original mp3\'s or the resulting word.csv\'s) that we\'re used to train the original thetas')
    parser.add_argument('-theta_file', type=str, nargs='*', help='the intial theta file(s)')
    parser.add_argument('-bitmask',   type=int, nargs='*', required=False, default=None,   metavar='splot_method_bitmask', 
                        help='Sets the methods used when converting the spectral plot from sox into the neural network\'s input')
    parser.add_argument('-hidden',    type=int, nargs='*', required=False, default=None, metavar='hidden_nodes', 
                        help='The number of hidden nodes to use in the neural networks')
    parser.add_argument('-iteration', type=int, nargs='*', required=False, default=None,  metavar='iterations', 
                        help='The number of iterations to train the neural network with')
    parser.add_argument('-ymap', type=str, nargs='*', required=False, default=None, metavar='y_mappings.txt', 
                        help='Supply a y_mappings.txt file')

    parser.add_argument('-funr', required=False, type=str, default="", metavar="filename", help='Tell the fuckit splitter to do noise reduction with the given file')
    parser.add_argument('-proxy_host', required=False, default="", type=str, metavar='proxy_host', help='Use the specified proxy host')
    parser.add_argument('-proxy_port', required=False, default="8118", type=str, metavar='proxy_host', help='Use the specified proxy port')
    parser.add_argument('-pp', required=False, action='store_true', help='Use the paypal downloader.')
    parser.add_argument('-t', required=False, action='store_true', help='Use tor.')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode.')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode.')
    args = parser.parse_args()

    if args.bitmask == None or args.bitmask == []:
        args.bitmask = [default_bitmask] * len(args.theta_file)
    if args.hidden == None or args.hidden == []:
        args.hidden = [default_hidden] * len(args.theta_file)
    if args.iteration == None or args.iteration == []:
        args.iteration = [default_iterations] * len(args.theta_file)
    if args.ymap == None or args.ymap == []:
        args.ymap = ["image-neural-network/y_mappings.txt"] * len(args.theta_file)

    if (len(args.bitmask) != len(args.theta_file) or len(args.hidden) != len(args.theta_file) or len(args.sample_dir) != len(args.theta_file) or
            len(args.iteration) != len(args.theta_file) or len(args.ymap) != len(args.theta_file)):
        print "The number of bitmask, hidden, iteration, samples dir, and ymap arguments must match the number of theta files specified (or be 0)"
        sys.exit()

    output_level = logging.CRITICAL
    if args.vv:
        output_level = logging.DEBUG
    elif args.v:
        output_level = logging.INFO

    downloaders = []
    filename_suffix = ".wav"
    downloaders.append(paypal_downloader.PaypalDownloader(args.proxy_host, args.proxy_port, output_level, args.t))

    splitters = []
    for i in range(0, 50):
        splitters.append(fuckit_splitter.FuckItSplitter(debug = output_level, filename_suffix = filename_suffix, noisered_file = args.funr))

    trainer = Trainer(output_level, downloaders, splitters, args.test_dir, args.download_dir, args.theta_file, args.ymap, args.sample_dir, 
                        args.bitmask, args.hidden, args.iteration, filename_suffix, args.funr)
    trainer.run()

